;; Generated package description from mode: lisp-data; .el  -*- denote-sequenceno-byte-compile: t -*-
(define-package "denote-sequence" "0.1.0" "Sequence notes or Folgezettel with Denote" '((emacs "28.1") (denote "3.1.0")) :commit "fd938da75573cc8b70e77ea00248c0c446613d43" :authors '(("Protesilaos Stavrou" . "info@protesilaos.com")) :maintainer '("Protesilaos Stavrou" . "info@protesilaos.com") :url "https://github.com/protesilaos/denote-sequence")
